import psycopg
from typing import Any, Callable

class PGDB:
    '''class for interacting with Postgres Databases\n
    conn_dict: dict = {
    \t "user" : "",
    \t "password" : "",
    \t "host" : "",
    \t "port" : "",
    \t "dbname" : ""
    }
    read_only (optional, default=True): Set the connection to be read only\n
    sql_dir (optional): Specify the absolute path of the directory where queries are stored\n
    README at https://github.com/HFxLhT8JqeU5BnUG/etb-pg'''

    def __init__(self, conn_dict: dict[str, str],
                 read_only: bool = True, sql_dir: str = ""):

        self.__connection__ = psycopg.connect(**conn_dict)
        self.__connection__.read_only = read_only

        if sql_dir and not sql_dir.endswith('/'):
            sql_dir += '/' 

        self.__SQL_DIR__ = sql_dir


    def execute_file_query(self, file_name: str):
        '''Executes the first query found in the specified file\n
        If no file extension is included in the file_name, it is assumed that it is a .sql file\n
        This will search the sql_dir set when object is instantiated. If no SQL directory was set,
        then it will assume that file_name is an absolute path'''

        if '.' not in file_name:
            file_name += '.sql'

        if self.__SQL_DIR__:
            location = self.__SQL_DIR__ + file_name
        else:
            location = file_name

        with open(location) as file:
            query = file.read()

        return self._execute_query(query)


    def execute_str_query(self, query: str):
        '''Execute a raw string query'''
        return self._execute_query(query)


    def _execute_query(self, query: str) -> list[dict[str, Any]] | None:
        '''Internal use, called by class methods that accept various input formats'''

        try:
            with self.__connection__.cursor() as curs:

                curs.execute(query)
                self.__connection__.commit()

                result = curs.fetchall()
                cols = curs.description

            rows = [{
                cols[i].name : row[i] for i in range(len(row))
            } for row in result]

            return rows

        except psycopg.ProgrammingError:
            return

        except Exception as e:
            print(e)
            return


    def str_to_query(self, fun: Callable[..., str]):
        '''Decorator for function that returns a string\n
        Will treat the string returned by decorated function as a query and return the result'''
        
        def wrapper(*args, **kwargs):
            query = fun(*args, **kwargs)
            assert isinstance(query, str), f'Decorated function must return a string, not {type(query)}'

            return self.execute_str_query(query)

        return wrapper


    def get_columns(self, table_name: str, schema: str = "public") -> dict[str, Any]:
        '''Pass the name of the table, and (optional) the schema\n
        Returns names and types of rows as dictionary'''
        location = schema + "." + table_name

        query = f"select * from {location} limit 1;"

        row = self._execute_query(query)[0]

        return {key: type(value) for key, value in row.items()}


    def get_rows(self, table_name: str, schema: str = "public", cols: list[str] = None):
        '''Select specified columns from a table\n
        schema (optional, defaults to public)\n
        cols (optional, defaults to "*")'''
        if not cols:
            select_cols = '*'
        else:
            select_cols = ', '.join(cols)

        location = schema + "." + table_name

        query = f"select {select_cols} from {location};"

        return self._execute_query(query)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['etb_pg']

package_data = \
{'': ['*']}

install_requires = \
['psycopg>=3.1.18,<4.0.0']

setup_kwargs = {
    'name': 'etb-pg',
    'version': '0.1.3',
    'description': 'Interface for PostgreSQL databases',
    'long_description': '# PG Library\n\nA small class for interacting with Postgres databases.\n\n## Initialize\n\nPass a dictionary with connection credentials when instantiating the PGDB object.\n```\nconn_dict: dict = {\n    "user" : "",        # default $USER\n    "password" : "",    # default None\n    "host" : "",        # default localhost\n    "port" : "",        # default 5432\n    "dbname" : ""       # default $USER\n}\n```\nAll of the connection params are optional, and default to the regular Postgres defaults (commented above).\n\nOptional args are read_only connection (default True) and an SQL directory where the object will check for queries stored as as readable files (default None).\n\n\n## Usage\n\nYou can execute a query from a file, from a raw string, from a decorated function, or run a very simple select query. As with the actual Postgres database, schema is always an optional argument, and always defaults to "public" if not provided. Results of a query are always returned as a list of dictionaries, in which the column name maps to the value.\n\nexecute_file_query(): If no SQL directory is passed when the object is created, then it will assume you\'re passing with an absolute file path. If there is no file extension included in the method call, then it will assume it\'s looking for a .sql file. If you did pass an SQL directory at instantiation, and you\'re using .sql file extensions, you can just pass the name of the file as a string without path or extension.\n\nexecute_str_query(): Execute a raw string query.\n\nstr_to_query(): Use this to decorate a function that returns a string. Could be useful for mildly dynamic queries. For example (assuming you have already created a PGDB object called "pg_interface"):\n\n```\n@pg_interface.str_to_query\ndef foo(cols: list[str], identifier: str) -> str:\n    columns = \', \'.join(cols)\n\n    query = f"""select {columns}\n            from schema.table_name st\n            where st.id = \'{identifier}\'"""\n    \n    return query\n```\nUsing the decorator, foo() will return the result of the query.\n\nget_columns(): Not a query function. Will return a list of {column_name : python_type} for the specified schema/table.\n\nget_rows(): Will return the results of a query that selects the specified columns from the specified schema/table.\n\n\nThere are no fancy methods for doing joins or window functions or creates/updates/deletes or anything like that. I want to keep any logic that\'s even remotely complex in a version controlled .sql file, or let the user wrap the logic in a decorated function. You can use the get_rows() method for very simple, unqualified selects from a table (or, preferably, a view or function).',
    'author': 'Tate Button',
    'author_email': 'yg3bpwn0or679hau8fxi@duck.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/HFxLhT8JqeU5BnUG/etb-pg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
